/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997-2004 Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_x11gl.c,v 1.17.2.4 2004/02/23 04:53:53 slouken Exp $";
#endif

#include <stdlib.h>	/* For getenv() prototype */
#include <string.h>

#include "SDL_events_c.h"
#include "SDL_error.h"
#include "SDL_x11video.h"
#include "SDL_x11dga_c.h"
#include "SDL_x11gl_c.h"

#define DEFAULT_OPENGL	"libGL.so.1"

XVisualInfo *X11_GL_GetVisual(_THIS)
{
#ifdef HAVE_OPENGL
	/* 64 seems nice. */
	int attribs[64];
	int i;

	/* load the gl driver from a default path */
	if ( ! this->gl_config.driver_loaded ) {
	        /* no driver has been loaded, use default (ourselves) */
	        if ( X11_GL_LoadLibrary(this, NULL) < 0 ) {
		        return NULL;
		}
	}

	/* See if we already have a window which we must use */
	if ( SDL_windowid ) {
		XWindowAttributes a;
		XVisualInfo vi_in;
		int out_count;

		XGetWindowAttributes(SDL_Display, SDL_Window, &a);
		vi_in.screen = SDL_Screen;
		vi_in.visualid = XVisualIDFromVisual(a.visual);
		glx_visualinfo = XGetVisualInfo(SDL_Display,
	                     VisualScreenMask|VisualIDMask, &vi_in, &out_count);
		return glx_visualinfo;
	}

        /* Setup our GLX attributes according to the gl_config. */
	i = 0;
	attribs[i++] = GLX_RGBA;
	attribs[i++] = GLX_RED_SIZE;
	attribs[i++] = this->gl_config.red_size;
	attribs[i++] = GLX_GREEN_SIZE;
	attribs[i++] = this->gl_config.green_size;
	attribs[i++] = GLX_BLUE_SIZE;
	attribs[i++] = this->gl_config.blue_size;

	if( this->gl_config.alpha_size ) {
		attribs[i++] = GLX_ALPHA_SIZE;
		attribs[i++] = this->gl_config.alpha_size;
	}

	if( this->gl_config.buffer_size ) {
		attribs[i++] = GLX_BUFFER_SIZE;
		attribs[i++] = this->gl_config.buffer_size;
	}

	if( this->gl_config.double_buffer ) {
		attribs[i++] = GLX_DOUBLEBUFFER;
	}

	attribs[i++] = GLX_DEPTH_SIZE;
	attribs[i++] = this->gl_config.depth_size;

	if( this->gl_config.stencil_size ) {
		attribs[i++] = GLX_STENCIL_SIZE;
		attribs[i++] = this->gl_config.stencil_size;
	}

	if( this->gl_config.accum_red_size ) {
		attribs[i++] = GLX_ACCUM_RED_SIZE;
		attribs[i++] = this->gl_config.accum_red_size;
	}

	if( this->gl_config.accum_green_size ) {
		attribs[i++] = GLX_ACCUM_GREEN_SIZE;
		attribs[i++] = this->gl_config.accum_green_size;
	}

	if( this->gl_config.accum_blue_size ) {
		attribs[i++] = GLX_ACCUM_BLUE_SIZE;
		attribs[i++] = this->gl_config.accum_blue_size;
	}

	if( this->gl_config.accum_alpha_size ) {
		attribs[i++] = GLX_ACCUM_ALPHA_SIZE;
		attribs[i++] = this->gl_config.accum_alpha_size;
	}

	if( this->gl_config.stereo ) {
		attribs[i++] = GLX_STEREO;
		attribs[i++] = this->gl_config.stereo;
	}
	
	if( this->gl_config.multisamplebuffers ) {
		attribs[i++] = GLX_SAMPLE_BUFFERS_ARB;
		attribs[i++] = this->gl_config.multisamplebuffers;
	}
	
	if( this->gl_config.multisamplesamples ) {
		attribs[i++] = GLX_SAMPLES_ARB;
		attribs[i++] = this->gl_config.multisamplesamples;
	}

#ifdef GLX_DIRECT_COLOR /* Try for a DirectColor visual for gamma support */
	if( !getenv("SDL_VIDEO_X11_NODIRECTCOLOR") ) {
		attribs[i++] = GLX_X_VISUAL_TYPE;
		attribs[i++] = GLX_DIRECT_COLOR;
	}
#endif
	attribs[i++] = None;

 	glx_visualinfo = this->gl_data->glXChooseVisual(GFX_Display, 
						  SDL_Screen, attribs);
#ifdef GLX_DIRECT_COLOR
	if( !glx_visualinfo && !getenv("SDL_VIDEO_X11_NODIRECTCOLOR") ) {
		/* No DirectColor visual?  Try again.. */
		attribs[i-3] = None;
 		glx_visualinfo = this->gl_data->glXChooseVisual(GFX_Display, 
						  SDL_Screen, attribs);
	}
#endif
	if( !glx_visualinfo ) {
		SDL_SetError( "Couldn't find matching GLX visual");
		return NULL;
	}
/*
	printf("Found GLX visual 0x%x\n", glx_visualinfo->visualid);
*/
	return glx_visualinfo;
#else
	SDL_SetError("X11 driver not configured with OpenGL");
	return NULL;
#endif
}

int X11_GL_CreateWindow(_THIS, int w, int h)
{
	int retval;
#ifdef HAVE_OPENGL
	XSetWindowAttributes attributes;
	unsigned long mask;
	unsigned long black;

	black = (glx_visualinfo->visual == DefaultVisual(SDL_Display,
						 	SDL_Screen))
	       	? BlackPixel(SDL_Display, SDL_Screen) : 0;
	attributes.background_pixel = black;
	attributes.border_pixel = black;
	attributes.colormap = SDL_XColorMap;
	mask = CWBackPixel | CWBorderPixel | CWColormap;

	SDL_Window = XCreateWindow(SDL_Display, WMwindow,
			0, 0, w, h, 0, glx_visualinfo->depth,
			InputOutput, glx_visualinfo->visual,
			mask, &attributes);
	if ( !SDL_Window ) {
		SDL_SetError("Could not create window");
		return -1;
	}
	retval = 0;
#else
	SDL_SetError("X11 driver not configured with OpenGL");
	retval = -1;
#endif
	return(retval);
}

int X11_GL_CreateContext(_THIS)
{
	int retval;
#ifdef HAVE_OPENGL
	/* We do this to create a clean separation between X and GLX errors. */
	XSync( SDL_Display, False );
	glx_context = this->gl_data->glXCreateContext(GFX_Display, 
				     glx_visualinfo, NULL, True);
	XSync( GFX_Display, False );

	if (glx_context == NULL) {
		SDL_SetError("Could not create GL context");
		return -1;
	}

	gl_active = 1;
	gl_extensions_checked = 0;
#else
	SDL_SetError("X11 driver not configured with OpenGL");
#endif
	if ( gl_active ) {
		retval = 0;
	} else {
		retval = -1;
	}
	return(retval);
}

void X11_GL_Shutdown(_THIS)
{
#ifdef HAVE_OPENGL
	/* Clean up OpenGL */
	if( glx_context ) {
		this->gl_data->glXMakeCurrent(GFX_Display, None, NULL);

		if (glx_context != NULL)
			this->gl_data->glXDestroyContext(GFX_Display, glx_context);

		if( this->gl_data->glXReleaseBuffersMESA ) {
		    this->gl_data->glXReleaseBuffersMESA(GFX_Display,SDL_Window);
		}
		glx_context = NULL;
	}
	gl_active = 0;
#endif /* HAVE_OPENGL */
}

#ifdef HAVE_OPENGL

static int ExtensionSupported(const char *extension, const char *extensions)
{
	const char *start;
	const char *where, *terminator;

	/* Extension names should not have spaces. */
	where = strchr(extension, ' ');
	if ( where || *extension == '\0' )
	      return 0;
	
	if ( ! extensions )
		return 0;

	/* It takes a bit of care to be fool-proof about parsing the
	 *      OpenGL extensions string. Don't be fooled by sub-strings,
	 *           etc. */
	
	start = extensions;
	
	for (;;)
	{
		where = strstr(start, extension);
		if (!where) break;
		
		terminator = where + strlen(extension);
		if (where == start || *(where - 1) == ' ')
	        if (*terminator == ' ' || *terminator == '\0') return 1;

		start = terminator;
	}
	
	return 0;
}

static void CheckExtensions(_THIS)
{
	const char *extensions;
	const char *glx_extensions;
	
	/* What extensions are available? */
	extensions = (const char *)this->glGetString(GL_EXTENSIONS);
	if ( this->gl_data->glXQueryExtensionsString ) {
		glx_extensions = (const char *)this->gl_data->glXQueryExtensionsString(GFX_Display, SDL_Screen);
	} else {
		glx_extensions = NULL;
	}

	/* 
	 * The context is now current, check for glXReleaseBuffersMESA() 
	 * extension. If extension is _not_ supported, destroy the pointer 
	 * (to make sure it will not be called in X11_GL_Shutdown() ).
	 * 
	 * DRI/Mesa drivers include glXReleaseBuffersMESA() in the libGL.so, 
	 * but there's no need to call it (is is only needed for some old 
	 * non-DRI drivers).
	 * 
	 * When using for example glew (http://glew.sf.net), dlsym() for
	 * glXReleaseBuffersMESA() returns the pointer from the glew library
	 * (namespace conflict).
	 *
	 * The glXReleaseBuffersMESA() pointer in the glew is NULL, if the 
	 * driver doesn't support this extension. So blindly calling it will
	 * cause segfault with DRI/Mesa drivers!
	 * 
	 */
	
	if ( ! ExtensionSupported("glXReleaseBuffersMESA", extensions) ) {
		this->gl_data->glXReleaseBuffersMESA = NULL;
	}

	if ( ((this->gl_data->glx_version >= MAKE_GLX_VERSION(1, 3)) ||
	      (ExtensionSupported("GLX_SGIX_fbconfig", glx_extensions) &&
	       ExtensionSupported("GLX_SGIX_pbuffer", glx_extensions))) &&
	     (this->gl_data->glXChooseFBConfigSGIX != NULL) &&
	     (this->gl_data->glXGetFBConfigAttribSGIX != NULL) &&
	     ((this->gl_data->glXCreateGLXPbufferSGIX != NULL) ||
	      (this->gl_data->glXCreatePbuffer != NULL)) &&
	     (this->gl_data->glXDestroyGLXPbufferSGIX != NULL) ) {
		this->GL_CreateRenderTarget = X11_GL_CreateRenderTarget;
		this->GL_GetRenderTargetAttribute = X11_GL_GetRenderTargetAttribute;
		this->GL_LockRenderTarget = X11_GL_LockRenderTarget;
		this->GL_UnlockRenderTarget = X11_GL_UnlockRenderTarget ;
		this->GL_FreeRenderTarget = X11_GL_FreeRenderTarget;
	} else {
		this->GL_CreateRenderTarget = NULL;
		this->GL_GetRenderTargetAttribute = NULL;
		this->GL_LockRenderTarget = NULL;
		this->GL_UnlockRenderTarget = NULL ;
		this->GL_FreeRenderTarget = NULL;
	}
}

/* Make the current context active */
int X11_GL_MakeCurrent(_THIS)
{
	int retval;
	
	retval = 0;
	if ( ! this->gl_data->glXMakeCurrent(GFX_Display,
	                                     SDL_Window, glx_context) ) {
		SDL_SetError("Unable to make GL context current");
		retval = -1;
	}
	XSync( GFX_Display, False );

	/* Once the context is current, check extensions */
	if ( ! gl_extensions_checked ) {
		CheckExtensions(this);
		gl_extensions_checked = 1;
	}

	/* More Voodoo X server workarounds... Grr... */
	SDL_Lock_EventThread();
	X11_CheckDGAMouse(this);
	SDL_Unlock_EventThread();

	return(retval);
}

/* Get attribute data from glX. */
int X11_GL_GetAttribute(_THIS, SDL_GLattr attrib, int* value)
{
	int retval;
	int glx_attrib = None;

	switch( attrib ) {
	    case SDL_GL_RED_SIZE:
		glx_attrib = GLX_RED_SIZE;
		break;
	    case SDL_GL_GREEN_SIZE:
		glx_attrib = GLX_GREEN_SIZE;
		break;
	    case SDL_GL_BLUE_SIZE:
		glx_attrib = GLX_BLUE_SIZE;
		break;
	    case SDL_GL_ALPHA_SIZE:
		glx_attrib = GLX_ALPHA_SIZE;
		break;
	    case SDL_GL_DOUBLEBUFFER:
		glx_attrib = GLX_DOUBLEBUFFER;
		break;
	    case SDL_GL_BUFFER_SIZE:
		glx_attrib = GLX_BUFFER_SIZE;
		break;
	    case SDL_GL_DEPTH_SIZE:
		glx_attrib = GLX_DEPTH_SIZE;
		break;
	    case SDL_GL_STENCIL_SIZE:
		glx_attrib = GLX_STENCIL_SIZE;
		break;
	    case SDL_GL_ACCUM_RED_SIZE:
		glx_attrib = GLX_ACCUM_RED_SIZE;
		break;
	    case SDL_GL_ACCUM_GREEN_SIZE:
		glx_attrib = GLX_ACCUM_GREEN_SIZE;
		break;
	    case SDL_GL_ACCUM_BLUE_SIZE:
		glx_attrib = GLX_ACCUM_BLUE_SIZE;
		break;
	    case SDL_GL_ACCUM_ALPHA_SIZE:
		glx_attrib = GLX_ACCUM_ALPHA_SIZE;
		break;
	    case SDL_GL_STEREO:
		glx_attrib = GLX_STEREO;
		break;
 	    case SDL_GL_MULTISAMPLEBUFFERS:
 		glx_attrib = GLX_SAMPLE_BUFFERS_ARB;
 		break;
 	    case SDL_GL_MULTISAMPLESAMPLES:
 		glx_attrib = GLX_SAMPLES_ARB;
 		break;
	    default:
		return(-1);
	}

	retval = this->gl_data->glXGetConfig(GFX_Display, glx_visualinfo, glx_attrib, value);

	return retval;
}

void X11_GL_SwapBuffers(_THIS)
{
	this->gl_data->glXSwapBuffers(GFX_Display, SDL_Window);
}

typedef struct X11_RenderTargetData {
	GLXFBConfigSGIX config;
	GLXPbuffer pbuffer;
} X11_RenderTargetData;

/* This is only used by X11_GL_CreateRenderTarget() */
static int XErrorFlag = 0;
static int HandleXError( Display *dpy, XErrorEvent *event )
{
	XErrorFlag = 1;
	return 0;
}

#ifdef DEBUG_FBCONFIG
void PrintFBConfigInfo(_THIS, GLXFBConfigSGIX fbConfig)
{
	int bufferSize, level, doubleBuffer, stereo, auxBuffers;
	int redSize, greenSize, blueSize, alphaSize;
	int depthSize, stencilSize;
	int accumRedSize, accumBlueSize, accumGreenSize, accumAlphaSize;
	int drawableType, renderType, xRenderable, xVisual, id;

	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_BUFFER_SIZE, &bufferSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_LEVEL, &level);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_DOUBLEBUFFER, &doubleBuffer);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_STEREO, &stereo);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_AUX_BUFFERS, &auxBuffers);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_RED_SIZE, &redSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_GREEN_SIZE, &greenSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_BLUE_SIZE, &blueSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_ALPHA_SIZE, &alphaSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_DEPTH_SIZE, &depthSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_STENCIL_SIZE, &stencilSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_ACCUM_RED_SIZE, &accumRedSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_ACCUM_GREEN_SIZE, &accumGreenSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_ACCUM_BLUE_SIZE, &accumBlueSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_ACCUM_ALPHA_SIZE, &accumAlphaSize);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_DRAWABLE_TYPE_SGIX, &drawableType);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_RENDER_TYPE_SGIX, &renderType);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_X_RENDERABLE_SGIX, &xRenderable);
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_X_VISUAL_TYPE_EXT, &xVisual);
	if (!xRenderable || !(drawableType & GLX_WINDOW_BIT_SGIX))
		xVisual = -1;
	this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, fbConfig, GLX_FBCONFIG_ID_SGIX, &id);

	printf("Id 0x%x\n", id);
	printf("  Buffer Size: %d\n", bufferSize);
	printf("  Level: %d\n", level);
	printf("  Double Buffer: %s\n", doubleBuffer ? "yes" : "no");
	printf("  Stereo: %s\n", stereo ? "yes" : "no");
	printf("  Aux Buffers: %d\n", auxBuffers);
	printf("  Red Size: %d\n", redSize);
	printf("  Green Size: %d\n", greenSize);
	printf("  Blue Size: %d\n", blueSize);
	printf("  Alpha Size: %d\n", alphaSize);
	printf("  Depth Size: %d\n", depthSize);
	printf("  Stencil Size: %d\n", stencilSize);
	printf("  Accum Red Size: %d\n", accumRedSize);
	printf("  Accum Green Size: %d\n", accumGreenSize);
	printf("  Accum Blue Size: %d\n", accumBlueSize);
	printf("  Accum Alpha Size: %d\n", accumAlphaSize);
	printf("  Drawable Types: ");
	if (drawableType & GLX_WINDOW_BIT_SGIX)  printf("Window ");
	if (drawableType & GLX_PIXMAP_BIT_SGIX)  printf("Pixmap ");
	if (drawableType & GLX_PBUFFER_BIT_SGIX)  printf("PBuffer");
	printf("\n");
	printf("  Render Types: ");
	if (renderType & GLX_RGBA_BIT_SGIX)  printf("RGBA ");
	if (renderType & GLX_COLOR_INDEX_BIT_SGIX)  printf("CI ");
	printf("\n");
	printf("  X Renderable: %s\n", xRenderable ? "yes" : "no");
}
#endif /* DEBUG_FBCONFIG */

int X11_GL_CreateRenderTarget(_THIS, SDL_RenderTarget *target, int width, int height, int *attrib)
{
	X11_RenderTargetData *render_target;
	int (*oldHandler)(Display*, XErrorEvent*);
	/* 64 seems nice. */
	int i, attribs[64];
	int pbAttribs[] = {
		GLX_LARGEST_PBUFFER_SGIX, False,
		GLX_PRESERVED_CONTENTS_SGIX, True,
		GLX_PBUFFER_WIDTH, target->w,
		GLX_PBUFFER_HEIGHT, target->h,
		None
	};
	int nConfigs;
	GLXFBConfigSGIX *fbConfigs;

	/* Allocate memory for data */
	render_target = (X11_RenderTargetData *)malloc(sizeof(*render_target));
	if ( ! render_target ) {
		SDL_OutOfMemory();
		return(-1);
	}
	render_target->pbuffer = None;
	target->gl_data = render_target;

	/* Catch X protocol errors with our own error handler */
	oldHandler = XSetErrorHandler( HandleXError );

        /* Setup our GLX attributes */
	i = 0;
	attribs[i++] = GLX_RENDER_TYPE_SGIX;
	attribs[i++] = GLX_RGBA_BIT_SGIX;
	attribs[i++] = GLX_DRAWABLE_TYPE_SGIX;
	attribs[i++] = GLX_PBUFFER_BIT_SGIX;
	attribs[i++] = GLX_RED_SIZE;
	attribs[i++] = attrib[2*SDL_GL_RED_SIZE+1];
	attribs[i++] = GLX_GREEN_SIZE;
	attribs[i++] = attrib[2*SDL_GL_GREEN_SIZE+1];
	attribs[i++] = GLX_BLUE_SIZE;
	attribs[i++] = attrib[2*SDL_GL_BLUE_SIZE+1];
	attribs[i++] = GLX_ALPHA_SIZE;
	attribs[i++] = attrib[2*SDL_GL_ALPHA_SIZE+1];
	attribs[i++] = GLX_BUFFER_SIZE;
	attribs[i++] = attrib[2*SDL_GL_BUFFER_SIZE+1];
	attribs[i++] = GLX_DOUBLEBUFFER;
	attribs[i++] = attrib[2*SDL_GL_DOUBLEBUFFER+1];
	attribs[i++] = GLX_DEPTH_SIZE;
	attribs[i++] = attrib[2*SDL_GL_DEPTH_SIZE+1];
	attribs[i++] = GLX_STENCIL_SIZE;
	attribs[i++] = attrib[2*SDL_GL_STENCIL_SIZE+1];
	attribs[i++] = GLX_ACCUM_RED_SIZE;
	attribs[i++] = attrib[2*SDL_GL_ACCUM_RED_SIZE+1];
	attribs[i++] = GLX_ACCUM_GREEN_SIZE;
	attribs[i++] = attrib[2*SDL_GL_ACCUM_GREEN_SIZE+1];
	attribs[i++] = GLX_ACCUM_BLUE_SIZE;
	attribs[i++] = attrib[2*SDL_GL_ACCUM_BLUE_SIZE+1];
	attribs[i++] = GLX_ACCUM_ALPHA_SIZE;
	attribs[i++] = attrib[2*SDL_GL_ACCUM_ALPHA_SIZE+1];
	attribs[i++] = GLX_STEREO;
	attribs[i++] = attrib[2*SDL_GL_STEREO+1];
	attribs[i++] = None;

	/* Loop through available configs trying to create a pbuffer */
	fbConfigs = this->gl_data->glXChooseFBConfigSGIX(GFX_Display, SDL_Screen, attribs, &nConfigs);
#ifdef DEBUG_FBCONFIG
	printf("%d configs found\n", nConfigs);
	for ( i = 0; i < nConfigs; ++i ) {
		PrintFBConfigInfo(this, fbConfigs[i]);
	}
#endif
	for ( i = 0; i < nConfigs; ++i ) {
		XErrorFlag = 0;
		if ( this->gl_data->glx_version >= MAKE_GLX_VERSION(1, 3) ) {
			render_target->pbuffer = this->gl_data->glXCreatePbuffer(GFX_Display, fbConfigs[i], pbAttribs);
		} else {
			render_target->pbuffer = this->gl_data->glXCreateGLXPbufferSGIX(GFX_Display, fbConfigs[i], target->w, target->h, pbAttribs);
		}
		if ( XErrorFlag ) {
			render_target->pbuffer = None;
		}
		if ( render_target->pbuffer != None ) {
			render_target->config = fbConfigs[i];
			break;
		}
	}

	/* Restore original X error handler */
	XSetErrorHandler( oldHandler );

	/* Return an error if we couldn't create a pbuffer */
	if ( render_target->pbuffer == None ) {
		SDL_SetError("Couldn't create pbuffer with desired attributes");
		return(-1);
	}
	return(0);
}

int X11_GL_GetRenderTargetAttribute(_THIS, SDL_RenderTarget *target, SDL_GLattr attrib, int* value)
{
	X11_RenderTargetData *render_target = (X11_RenderTargetData *)target->gl_data;
	int retval;
	int glx_attrib = None;

	switch( attrib ) {
	    case SDL_GL_RED_SIZE:
		glx_attrib = GLX_RED_SIZE;
		break;
	    case SDL_GL_GREEN_SIZE:
		glx_attrib = GLX_GREEN_SIZE;
		break;
	    case SDL_GL_BLUE_SIZE:
		glx_attrib = GLX_BLUE_SIZE;
		break;
	    case SDL_GL_ALPHA_SIZE:
		glx_attrib = GLX_ALPHA_SIZE;
		break;
	    case SDL_GL_DOUBLEBUFFER:
		glx_attrib = GLX_DOUBLEBUFFER;
		break;
	    case SDL_GL_BUFFER_SIZE:
		glx_attrib = GLX_BUFFER_SIZE;
		break;
	    case SDL_GL_DEPTH_SIZE:
		glx_attrib = GLX_DEPTH_SIZE;
		break;
	    case SDL_GL_STENCIL_SIZE:
		glx_attrib = GLX_STENCIL_SIZE;
		break;
	    case SDL_GL_ACCUM_RED_SIZE:
		glx_attrib = GLX_ACCUM_RED_SIZE;
		break;
	    case SDL_GL_ACCUM_GREEN_SIZE:
		glx_attrib = GLX_ACCUM_GREEN_SIZE;
		break;
	    case SDL_GL_ACCUM_BLUE_SIZE:
		glx_attrib = GLX_ACCUM_BLUE_SIZE;
		break;
	    case SDL_GL_ACCUM_ALPHA_SIZE:
		glx_attrib = GLX_ACCUM_ALPHA_SIZE;
		break;
	    case SDL_GL_STEREO:
		glx_attrib = GLX_STEREO;
		break;
	    default:
		return(-1);
	}

	retval = this->gl_data->glXGetFBConfigAttribSGIX(GFX_Display, render_target->config, glx_attrib, value);

	return retval;
}

void X11_GL_LockRenderTarget(_THIS, SDL_RenderTarget *target)
{
	X11_RenderTargetData *render_target = (X11_RenderTargetData *)target->gl_data;
	/* Set the draw target */
	this->gl_data->glXMakeCurrent(GFX_Display, render_target->pbuffer, glx_context);
	XSync( GFX_Display, False );
}

void X11_GL_UnlockRenderTarget(_THIS, SDL_RenderTarget *target)
{
	/* Update the texture with the pbuffer contents */
	if ( target->texture ) {
		this->glBindTexture(GL_TEXTURE_2D, target->texture);
		this->glCopyTexSubImage2D(GL_TEXTURE_2D,
					  0,
					  0, 0,
					  0, 0,
					  target->w, target->h);
	}

	/* Reset the draw target */
	this->gl_data->glXMakeCurrent(GFX_Display, SDL_Window, glx_context);
	XSync( GFX_Display, False );
}

void X11_GL_FreeRenderTarget(_THIS, SDL_RenderTarget *target)
{
	X11_RenderTargetData *render_target = (X11_RenderTargetData *)target->gl_data;
	if ( render_target->pbuffer != None ) {
		this->gl_data->glXDestroyGLXPbufferSGIX(GFX_Display, render_target->pbuffer);
	}
	free(render_target);
}

#endif /* HAVE_OPENGL */

void X11_GL_UnloadLibrary(_THIS)
{
#ifdef HAVE_OPENGL
	if ( this->gl_config.driver_loaded ) {
		dlclose(this->gl_config.dll_handle);

		this->gl_data->glXGetProcAddress = NULL;
		this->gl_data->glXChooseVisual = NULL;
		this->gl_data->glXCreateContext = NULL;
		this->gl_data->glXDestroyContext = NULL;
		this->gl_data->glXMakeCurrent = NULL;
		this->gl_data->glXSwapBuffers = NULL;

		this->gl_data->glXChooseFBConfigSGIX = NULL;
		this->gl_data->glXGetFBConfigAttribSGIX = NULL;
		this->gl_data->glXCreateGLXPbufferSGIX = NULL;
		this->gl_data->glXDestroyGLXPbufferSGIX = NULL;
		this->gl_data->glXCreatePbuffer = NULL;

		this->gl_config.dll_handle = NULL;
		this->gl_config.driver_loaded = 0;
	}
#endif
}

#ifdef HAVE_OPENGL

static int GetGLXVersion(_THIS, void *handle)
{
	int glx_major_version, glx_minor_version;

	/* The ATI drivers put the correct version in the client string */
	const char *version_string;
	const char *(*glXGetClientString)(Display *dpy, int name) =
		(const char *(*)(Display *, int)) dlsym(handle, "glXGetClientString");
	Bool (*glXQueryVersion)(Display *dpy, int *major, int *minor) =
		(Bool (*)(Display *, int *, int *)) dlsym(handle, "glXQueryVersion");

	if ( glXGetClientString &&
	     (version_string=glXGetClientString(GFX_Display, GLX_VERSION)) &&
	     sscanf(version_string, "%d.%d", 
	            &glx_major_version,
	            &glx_minor_version) ) {
		return MAKE_GLX_VERSION(glx_major_version, glx_minor_version);
	}
	if ( glXQueryVersion &&
	     glXQueryVersion(GFX_Display,
	                     &glx_major_version,
	                     &glx_minor_version) ) {
		return MAKE_GLX_VERSION(glx_major_version, glx_minor_version);
	}
	return 0;
}

/* Passing a NULL path means load pointers from the application */
int X11_GL_LoadLibrary(_THIS, const char* path) 
{
	void* handle;
	int dlopen_flags;

 	if ( gl_active ) {
 		SDL_SetError("OpenGL context already created");
 		return -1;
 	}

#ifdef RTLD_GLOBAL
	dlopen_flags = RTLD_LAZY | RTLD_GLOBAL;
#else
	dlopen_flags = RTLD_LAZY;
#endif
	handle = dlopen(path, dlopen_flags);
	/* Catch the case where the application isn't linked with GL */
	if ( (dlsym(handle, "glXChooseVisual") == NULL) && (path == NULL) ) {
		dlclose(handle);
		path = getenv("SDL_VIDEO_GL_DRIVER");
		if ( path == NULL ) {
			path = DEFAULT_OPENGL;
		}
		handle = dlopen(path, dlopen_flags);
	}
	if ( handle == NULL ) {
		SDL_SetError("Could not load OpenGL library");
		return -1;
	}

	/* Unload the old driver and reset the pointers */
	X11_GL_UnloadLibrary(this);

	/* Try to figure out the GLX version */
	this->gl_data->glx_version = GetGLXVersion(this, handle);

	/* Load new function pointers */
	this->gl_data->glXGetProcAddress =
		(void *(*)(const GLubyte *)) dlsym(handle, "glXGetProcAddressARB");
	this->gl_data->glXChooseVisual =
		(XVisualInfo *(*)(Display *, int, int *)) dlsym(handle, "glXChooseVisual");
	this->gl_data->glXCreateContext =
		(GLXContext (*)(Display *, XVisualInfo *, GLXContext, int)) dlsym(handle, "glXCreateContext");
	this->gl_data->glXDestroyContext =
		(void (*)(Display *, GLXContext)) dlsym(handle, "glXDestroyContext");
	this->gl_data->glXMakeCurrent =
		(int (*)(Display *, GLXDrawable, GLXContext)) dlsym(handle, "glXMakeCurrent");
	this->gl_data->glXSwapBuffers =
		(void (*)(Display *, GLXDrawable)) dlsym(handle, "glXSwapBuffers");
	this->gl_data->glXGetConfig =
		(int (*)(Display *, XVisualInfo *, int, int *)) dlsym(handle, "glXGetConfig");
	this->gl_data->glXQueryExtensionsString =
		(const char *(*)(Display *, int)) dlsym(handle, "glXQueryExtensionsString");

	/* We don't compare below for this in case we're not using Mesa. */
	this->gl_data->glXReleaseBuffersMESA =
		(void (*)(Display *, GLXDrawable)) dlsym( handle, "glXReleaseBuffersMESA" );
	
	/* We don't compare below for this in case we don't have pbuffer extensions */
	if ( this->gl_data->glx_version >= MAKE_GLX_VERSION(1, 3) ) {
		this->gl_data->glXChooseFBConfigSGIX =
			(GLXFBConfigSGIX *(*)(Display*, int, int*, int*)) dlsym( handle, "glXChooseFBConfig" );
		this->gl_data->glXGetFBConfigAttribSGIX =
			(int (*)(Display*, GLXFBConfigSGIX, int, int*)) dlsym( handle, "glXGetFBConfigAttrib" );
		this->gl_data->glXCreatePbuffer =
			(GLXPbuffer (*)(Display*, GLXFBConfigSGIX, int*)) dlsym( handle, "glXCreatePbuffer" );
		this->gl_data->glXDestroyGLXPbufferSGIX =
			(void (*)(Display*, GLXPbuffer)) dlsym( handle, "glXDestroyPbuffer" );
	} else {
		this->gl_data->glXChooseFBConfigSGIX =
			(GLXFBConfigSGIX *(*)(Display*, int, int*, int*)) dlsym( handle, "glXChooseFBConfigSGIX" );
		this->gl_data->glXGetFBConfigAttribSGIX =
			(int (*)(Display*, GLXFBConfigSGIX, int, int*)) dlsym( handle, "glXGetFBConfigAttribSGIX" );
		this->gl_data->glXCreateGLXPbufferSGIX =
			(GLXPbuffer (*)(Display*, GLXFBConfigSGIX, unsigned int, unsigned int, int*)) dlsym( handle, "glXCreateGLXPbufferSGIX" );
		this->gl_data->glXDestroyGLXPbufferSGIX =
			(void (*)(Display*, GLXPbuffer)) dlsym( handle, "glXDestroyGLXPbufferSGIX" );
	}

	if ( (this->gl_data->glXChooseVisual == NULL) || 
	     (this->gl_data->glXCreateContext == NULL) ||
	     (this->gl_data->glXDestroyContext == NULL) ||
	     (this->gl_data->glXMakeCurrent == NULL) ||
	     (this->gl_data->glXSwapBuffers == NULL) ||
	     (this->gl_data->glXGetConfig == NULL) ||
	     (this->gl_data->glXQueryExtensionsString == NULL)) {
		SDL_SetError("Could not retrieve OpenGL functions");
		return -1;
	}

	this->gl_config.dll_handle = handle;
	this->gl_config.driver_loaded = 1;
	if ( path ) {
		strncpy(this->gl_config.driver_path, path,
			sizeof(this->gl_config.driver_path)-1);
	} else {
		strcpy(this->gl_config.driver_path, "");
	}
	return 0;
}

void *X11_GL_GetProcAddress(_THIS, const char* proc)
{
	static char procname[1024];
	void* handle;
	void* retval;
	
	handle = this->gl_config.dll_handle;
	if ( this->gl_data->glXGetProcAddress ) {
		return this->gl_data->glXGetProcAddress(proc);
	}
#if defined(__OpenBSD__) && !defined(__ELF__)
#undef dlsym(x,y);
#endif
	retval = dlsym(handle, proc);
	if (!retval && strlen(proc) <= 1022) {
		procname[0] = '_';
		strcpy(procname + 1, proc);
		retval = dlsym(handle, procname);
	}
	return retval;
}

#endif /* HAVE_OPENGL */
